/*
 *  YASS 1.10
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* 1) include utils macro */
#include "util.h"
/* 2) include global variables */
#include "global_var.h"
/* 3) the current file defs */
#include "assemble.h"

/* 4) other files */
#include "kword.h"
#include "tuple.h"
#include "prdyn.h"
#include "proba.h"
#include "align.h"
#include "display.h"
#include "hash.h"
 
/*
 * Hach table division
 */
#define DIV_HACH 135
#define H_DIAG(diagonal) ((diagonal)/DIV_HACH)
#define LAST_CHUNK_DIAG_NB 65535


/*
 * Modulo access to the last tuple
 */

#ifdef MEMOPT
  /* query table index (little more time consuming but memory economic ) */
  #define MEMOPT_SETMOD   { int i=0; memopt_modmask = 0; \
	for (i=0; (                                      \
                   (ALIGN_EVERY_NB_ITER + 1 +            \
                    keysize_query       + 1 +            \
                    gp_rho_stat         + 1 +            \
                    gp_delta_stat       + 1              \
                    )                                    \
                    >= memopt_modmask)                   \
                  ;i++)                                  \
                 memopt_modmask |=  (1 << i);            \
  }

  #define MEMOPT_MOD      &(memopt_modmask)
  #define MEMOPT_TABSIZE   (memopt_modmask+1)
#else
  /* full table index (memory consuming) */ 
  #define MEMOPT_SETMOD
  #define MEMOPT_MOD       
  #define MEMOPT_TABSIZE   (keysize_query + keysize_text)
#endif

/*
 * Shift tables used in assemble algorithms
 */


int initialise_deltashift()
{
  int i,s=1;
  if (gv_delta_shift == NULL){
    gv_delta_shift = (int * ) MALLOC ((1+2*gp_delta_stat)*sizeof(int));
    ASSERT(gv_delta_shift,"initialise_deltashift");
    
    for ( i=0 ; i<1+2*gp_delta_stat ; i++) {
      gv_delta_shift[i] = s * (i+1) / 2;
      s *= -1;
    }
  }
  return 0;
}

/*
 *  definition des fonctions de la table hash ou tableaux
 */
#ifdef LOW_MEMORY
#define INIT_TAB(table,type,size,set)     table=initHash(gp_hash_size,gp_hash_element)
#define GET_TAB_MIN(table,index,minValue) getHashValueMin(&table,index,minValue)
#define GET_TAB(table,index)              getHashValue(&table,index)
#define EXIST_TAB(table,index)            existHashDiag(&table,index)
#define PUT_TAB(table,index,value)        putHashOneDiagValue (&table,index,(int)value)
#define DEL_TAB(table,index,value)        delHash (&table,index)
#define FREE_TAB(table)                   freeHash(&table)
#define RESET_TAB(table,size,set)         resetHash(&table);
#else
#define INIT_TAB(table,type,size,set)    {                                    \
     table =  (type) MALLOC(size);                                            \
     ASSERT(table,errmessage);                                                \
     memset(table,set,(unsigned)(size));                                      \
}

#define GET_TAB_MIN(table,index,minValue) ((table)[(index) MEMOPT_MOD])
#define GET_TAB(table,index)              ((table)[(index) MEMOPT_MOD])
#define EXIST_TAB(table,index)            ((table)[(index) MEMOPT_MOD])
#define PUT_TAB(table,index,value)        ((table)[(index) MEMOPT_MOD] = value)
#define DEL_TAB(table,index,value)        ((table)[(index) MEMOPT_MOD] = value)
#define FREE_TAB(table)                   {FREE(table,MEMOPT_TABSIZE * sizeof(int));}
#define RESET_TAB(table,size,set)         {memset(table,set,(unsigned)(size));}
#endif










int Assemble_Single(/*in*/ char * data, /*in*/ int datasize, Feature * f)
{
   
  /* main variables 
   * for the linked list computation
   */
  int * keylist[MAX_SEED];
  int   keylist_size[MAX_SEED];
  int * first[MAX_SEED];
  int   first_size[MAX_SEED];
  int   keysize_query = datasize - gp_seeds_span_min + 1;
#ifdef MEMOPT
  int   memopt_modmask = 0;
#else
  int   keysize_text   = 0;
#endif
  register int index=0;
  register int index_fin=0;
  
  /* for the linking algorithm */
  int i_previous;
  int j;
  int diagonal,observed_diagonal,delta_diagonal;
   
  /* auxiliary variables */
  int code;
#ifdef LOW_MEMORY
  /* use hash table*/
  Table_hash  last_tuple_pos_with_diag;
  Table_hash  last_tuple_ptr_with_diag;
#else
  int  *   last_tuple_pos_with_diag = NULL;
  tuple ** last_tuple_ptr_with_diag = NULL;
#endif

  /* For multi seed */
  int seed;

  /* Cache */
#ifdef CACHE
  int j_hack,bcode; 
  register int borne_min_hash=gp_rho_stat/2-1;
  register int borne_max_hash=DIV_HACH-gp_rho_stat/2+1;
  int *last_tuple_pos_with_diag_hach;
#endif

  
  /* sequence covering variables */ 
  tuple     * t, * t_;
    
  /* [0] left correction factor set to zero */
  f->left_correction =  0;

  /* to keep lists of tuple found */
  f->first_tl    = CreateTupleList();
  f->last_tl     = f->first_tl;

  /* [1] key linked list creation */
#ifdef STATS
  last_clock = clock();
#endif
  for( seed = 0 ; seed < gp_nb_seeds ; seed ++ ) 
    CreateKeyList(data,datasize,seed,&first[seed],&first_size[seed],&keylist[seed],&keylist_size[seed],f);


  /* [2] allocations */
  INIT_TAB(last_tuple_pos_with_diag,int*,MEMOPT_TABSIZE*sizeof(int),0xff);
  INIT_TAB(last_tuple_ptr_with_diag,tuple**,MEMOPT_TABSIZE*sizeof(tuple*),0x00);

  

#ifdef CACHE
  last_tuple_pos_with_diag_hach = (int *) MALLOC((unsigned)(((MEMOPT_TABSIZE)/DIV_HACH+1) * sizeof(int)));
  ASSERT(last_tuple_pos_with_diag_hach,"Assemble");
  memset(last_tuple_pos_with_diag_hach,0x0,(unsigned)(((MEMOPT_TABSIZE)/DIV_HACH+1) * sizeof(int)));
#endif

  
  
  
  /* [3] algorithm to create linked tuples */
#ifdef STATS
  f->clock_pre += clock() - last_clock;
  last_clock = clock();
#endif

  /* for each pos */
  for(f->i_current = 0 ; f->i_current <  datasize; f->i_current++ ) {

    /* for each seed */
    for( seed= 0 ; seed < gp_nb_seeds ; seed ++ ) {
      
      if ( f->i_current < datasize - gp_seeds_span[seed] + 1) {
	
	KEY(code,data,f->i_current,seed); /* we take the code of the kmot 'w' located at i_current position     */

#ifdef KEYLISTCOMPRESS	
	index      = first[seed][code];  /* search the first occurrence of 'w' on the sequence */
	index_fin  = first[seed][code+1];
	i_previous = keylist[seed][index];
#else	
	index      = first[seed][code];  /* search the first occurrence of 'w' on the sequence */
	index_fin  = -1;
	i_previous = index;
#endif
	/* consider all the occurrences of 'w' before i_current */
	while(i_previous < f->i_current - gp_distdiag && index != index_fin ){
	  diagonal = f->i_current - i_previous;
	  /* consider all the interesting diagonals by adding indel bias  0 +1 -1 +2... until we reach a new tuple or the upper indel bound */
	  observed_diagonal = diagonal;
	  delta_diagonal = 0;
	  
#ifdef CACHE 
	  /* bcode is a hash of */
	  j_hack = last_tuple_pos_with_diag_hach [H_DIAG(observed_diagonal )]; 
	  bcode  = ((j_hack < f->i_current - gp_rho_stat) || (j_hack > f->i_current));
	  if ( observed_diagonal % DIV_HACH > borne_max_hash ) {
	    j_hack = last_tuple_pos_with_diag_hach [H_DIAG(observed_diagonal)+1]; 
	    bcode  = bcode && ((j_hack< f->i_current - gp_rho_stat)||(j_hack > f->i_current)); 
	  }
	  if (observed_diagonal % DIV_HACH < borne_min_hash) {
	    j_hack = last_tuple_pos_with_diag_hach [H_DIAG(observed_diagonal)-1]; 
	    bcode = bcode && ((j_hack< f->i_current - gp_rho_stat)||(j_hack > f->i_current));
	  } 
	  if (bcode==0)
#endif
	    
	    while( delta_diagonal < 2 * gp_delta_stat + 1){
	      
	      /* search for interesting tuples */	  
	      j = GET_TAB_MIN(last_tuple_pos_with_diag,observed_diagonal,f->i_current - gp_rho_stat);
	      
	      /* if tuples found are respecting criteria (neighboors on diagonal and on sequences) */
	      if ( j > f->i_current - gp_rho_stat                   &&
		   j <= f->i_current                                &&
		   j - observed_diagonal <= f->i_current - diagonal &&
		   j != -1 ){
		
		
		/* hachDisplay (&last_tuple_ptr_with_diag); */
		/*---------------------------------------------------------------------*/
		
		/* (1) possibly create one instance of tuple [j]
		 * if this one was not in any list.
		 */
		if (EXIST_TAB(last_tuple_ptr_with_diag,observed_diagonal) == 0){
		  
#ifdef STATS
		  f->nb_chains_built++;
#endif
		  
		  /* create at tuple list element */
		  f->last_tl->next = CreateTupleList();
		  f->last_tl = f->last_tl->next;
		  
		  /* create a tuple [j] and update lists */
		  t  = CreateTuple(j,observed_diagonal,seed);
		  t->bitweight=gp_automata[gp_seeds_bitweight[seed]];
		  t->overlappedhits=gp_seeds_bitweight[seed];
		  f->last_tl->first_tuple = t;
		  PUT_TAB(last_tuple_ptr_with_diag,observed_diagonal,t);
		}
		else
		  /* (2) we search for the last tuple [j] instanciation
		   */
		  t = (tuple *)GET_TAB(last_tuple_ptr_with_diag,observed_diagonal) ;
		
		/*
		 * (3) iterate tuples and add this new one to the most adapted one
		 */
		
		do{
		  
		  if (  diagonal == t->diagonal  && f->i_current - t->occurrence <= gp_seeds_span[seed] + 1){
		    
		    /* (3a) possible fusion with this tuple (no instanciation)
		     *
		     * [inter-tuple overlap]
		     */ 
		    
		    t->leftsize          +=  f->i_current -  t->occurrence;
		    t->lastSeed           = seed;
		    t->bitweight         += gp_automata[(~(t->overlappedhits<<(f->i_current -  t->occurrence)))&gp_seeds_bitweight[t->lastSeed]];
		    t->overlappedhits     = t->overlappedhits<<(f->i_current -  t->occurrence) | gp_seeds_bitweight[t->lastSeed];
		    t->overlappedhits    &= AUTOMATAMASK; 
		    t->occurrence         = f->i_current;
		    goto no_maj_last_tuple_ptr_with_diag;
		    
		  }
		  
		  /* list search for the most fitted */
		  if(!(t->next))
		    break;
		  t = t->next;
		  
		}while(1);
		
		/*
		 * (3d) any fusion has not been possible
		 *
		 * we instanciate the tuple and add this one at the end of the chain.
		 *
		 */
		t_ = CreateTuple(f->i_current, diagonal, seed);
		t->next = t_;
		PUT_TAB(last_tuple_ptr_with_diag,diagonal,t_);
		
		/*
		 * (4) dont reset last_tuple_ptr_with_diag[diagonal] to NULL
		 * because it has been modified in (3)
		 */
		goto no_maj_last_tuple_ptr_with_diag;	  
	      }
	      
	      /*---------------------------------------------------------------------*/
	      /* diagonal walk d,d+1d-1,... */
	      delta_diagonal++;
	      observed_diagonal = diagonal + gv_delta_shift[delta_diagonal];
	      observed_diagonal = MAX(MIN(observed_diagonal,keysize_query-1),1);
	      /* border effects avoided*/
	      
	    }/* while : diagonal walk */
	  
	  
	  /* reset last tuple instancied to NULL (because not instance is created)
	   */
	  
	  DEL_TAB(last_tuple_ptr_with_diag,diagonal,NULL);	
	  
	  
	no_maj_last_tuple_ptr_with_diag:
	  
	  /*
	   * update last tuple position  with diagonal = 'diagonal'
	   */
#ifdef STATS
	  f->nb_seeds++;
#endif
	  PUT_TAB(last_tuple_pos_with_diag,diagonal,f->i_current);
	  
#ifdef CACHE
	  last_tuple_pos_with_diag_hach[H_DIAG(diagonal)]=f->i_current;
#endif
	  /* iterate 'w' code list before f->i_current position */
#ifdef KEYLISTCOMPRESS		
	  index ++;
	  i_previous = keylist[seed][index]; 	
#else	
	  i_previous = index = keylist[seed][index];
#endif
        }  
      }
    } /* for each seed*/
    
      
    if (!(f->i_current % ALIGN_EVERY_NB_ITER)){
#ifdef TRACE 
      Display_Progress((unsigned)f->i_current,(unsigned)(keysize_query - 1),f);
#endif 
	
#ifdef STATS
      f->clock_chain += clock() - last_clock;
      last_clock=  clock();
#endif		
#ifdef DEBUG_ASSEMBLE
      DisplayListTupleList(f->first_tl);
#endif
      AlignAndFree( 
		   data, datasize,
		   data, datasize,
		   0,
		   f
		   );
	
#ifdef STATS
      f->clock_align += clock() - last_clock;
      last_clock=  clock();
#endif
    }  
  } /* end of [3] */
  

#ifdef TRACE
  Display_Progress((unsigned)(unsigned)f->i_current,keysize_query - 1,f);
#endif  
  
#ifdef STATS
  f->clock_chain += clock() - last_clock;
  last_clock=  clock();
#endif	
  
  /* on aligne les tuples restants */
#ifdef DEBUG_ASSEMBLE
  DisplayListTupleList(f->first_tl);
#endif
  AlignAndFree(
	       data, datasize,
	       data, datasize,
	       1,
	       f
	       );
#ifdef STATS	
  f->clock_align += clock() - last_clock;
  last_clock =  clock();
#endif

#ifdef CACHE
 FREE(last_tuple_pos_with_diag_hach ,((MEMOPT_TABSIZE)/DIV_HACH+1) * sizeof(int));
#endif

  FREE_TAB(last_tuple_pos_with_diag);
  FREE_TAB(last_tuple_ptr_with_diag);
  for( seed = 0 ; seed < gp_nb_seeds ; seed ++ ) {
    FREE(keylist[seed], keylist_size[seed]);
    FREE(first[seed],   first_size[seed]);
  }
  return  0;
}


int Assemble_SingleRev(/*in*/   char * datarev, /*in*/  char * data, /*in*/ int datasize,
		       Feature  *f) {
  
  
  
  /* main variables */
  /* for the linked list computation */
  int * keylist[MAX_SEED];
  int   keylist_size[MAX_SEED];
  int * first[MAX_SEED];
  int   first_size[MAX_SEED];
  
  int keysize_query  = datasize - gp_seeds_span_min + 1;
#ifndef MEMOPT	
  int keysize_text   = keysize_query;
#else
  int memopt_modmask = 0;
#endif 
  
  /* for the linking algorithm */
  int index,index_fin;
  int i_previous;
  int j;
  int diagonal,observed_diagonal,delta_diagonal;
  
  /* auxiliary variables */
  int code;
#ifdef LOW_MEMORY
  /* use hash table*/
  Table_hash  last_tuple_pos_with_diag;
  Table_hash  last_tuple_ptr_with_diag;
#else
  int  *   last_tuple_pos_with_diag = NULL;
  tuple ** last_tuple_ptr_with_diag = NULL;
#endif
  
  /* For multi seed */
  int seed;
  
  /* Cache */
#ifdef CACHE
  int j_hack,bcode; 
  register int borne_min_hash=gp_rho_stat/2-1;
  register int borne_max_hash=DIV_HACH-gp_rho_stat/2+1;
  int *last_tuple_pos_with_diag_hach;
#endif
  
  /* sequence covering variables */ 
  tuple     * t, * t_;
  
  /* to keep lists of tuple found */
  f->first_tl = CreateTupleList();
  f->last_tl  = f->first_tl;
  
  
  /* left correction factor set to keysize_query */
  f->left_correction =  keysize_query;
  
  MEMOPT_SETMOD;  
  
  /* [1] key linked list creation */
#ifdef STATS
  last_clock = clock();
#endif
  for (seed=0;seed<gp_nb_seeds;seed++)
    CreateKeyList(datarev,datasize,seed,&first[seed],&first_size[seed],&keylist[seed],&keylist_size[seed],f);
  
  /* [2] allocations */
  INIT_TAB(last_tuple_pos_with_diag,int *,MEMOPT_TABSIZE * sizeof(int),0xff);
  INIT_TAB(last_tuple_ptr_with_diag,tuple **,MEMOPT_TABSIZE * sizeof(int),0x00);
  
#ifdef CACHE
  last_tuple_pos_with_diag_hach = (int *) MALLOC((unsigned)(((MEMOPT_TABSIZE)/DIV_HACH+1) * sizeof(int)));
  ASSERT(last_tuple_pos_with_diag_hach,"Assemble");
  memset(last_tuple_pos_with_diag_hach,0x0,(unsigned)(((MEMOPT_TABSIZE)/DIV_HACH+1) * sizeof(int)));
#endif
  
  /* [3] algorithm to create linked tuples */
#ifdef STATS
  f->clock_pre += clock() - last_clock;
  last_clock = clock();
#endif
  
  /* for each pos */
  for(f->i_current = 0 ; f->i_current < datasize ; f->i_current++ ) {
    /* for each seed */
    for ( seed = 0 ; seed < gp_nb_seeds ; seed++ ) {
      if ( f->i_current < datasize - gp_seeds_span[seed] + 1) {
	
	KEY(code,data,f->i_current,seed); /* we take the code of the kmot 'w' located at f->i_current position     */

#ifdef KEYLISTCOMPRESS	
	index      = first[seed][code];  /* search the first occurrence of 'w' on the sequence */
	index_fin  = first[seed][code+1];
	i_previous = keylist[seed][index];
#else	
	index      = first[seed][code];  /* search the first occurrence of 'w' on the sequence */
	index_fin  = -1;
	i_previous = index;
#endif
	
	/* consider all the occurrences of 'w' before f->i_current */
	
	
	while( i_previous < keysize_query - f->i_current  && index != index_fin){  
	  
	  diagonal =   f->i_current - i_previous + keysize_query ;      
	  /* consider all the interesting diagonals by adding indel bias  0 +1 -1 +2... until we reach a new tuple or the upper indel bound */
	  observed_diagonal = diagonal;
	  delta_diagonal = 0;
	  
#ifdef CACHE 
	  /* bcode is a hash of diags*/
	  j_hack = last_tuple_pos_with_diag_hach [H_DIAG(observed_diagonal )]; 
	  bcode  = ((j_hack < f->i_current - gp_rho_stat) || (j_hack > f->i_current));
	  if ( observed_diagonal % DIV_HACH > borne_max_hash ) {
	    j_hack = last_tuple_pos_with_diag_hach [H_DIAG(observed_diagonal)+1]; 
	    bcode  = bcode && ((j_hack< f->i_current - gp_rho_stat)||(j_hack > f->i_current)); 
	  }
	  if (observed_diagonal % DIV_HACH < borne_min_hash) {
	    j_hack = last_tuple_pos_with_diag_hach [H_DIAG(observed_diagonal)-1]; 
	    bcode = bcode && ((j_hack< f->i_current - gp_rho_stat)||(j_hack > f->i_current));
	  } 
	  if (bcode==0)
#endif	    
	    while( delta_diagonal < 2 * gp_delta_stat + 1){
	      
	      /* search for interesting tuples */
	      j = GET_TAB_MIN(last_tuple_pos_with_diag,observed_diagonal,f->i_current - gp_rho_stat);
	
	      
      
	      /* if tuples found are respecting criteria (neighboors on diagonal and on sequences) */
	      if ( j > f->i_current - gp_rho_stat                  &&                                         
		   j <= f->i_current                                && 
		   j - observed_diagonal <= f->i_current - diagonal && 
		   j != -1 ){ 
		
		
		/*---------------------------------------------------------------------*/
		
		/* (1) possibly create one instance of tuple [j] 
		 * if this one was not in any list. 
		 */
		
		if (EXIST_TAB(last_tuple_ptr_with_diag,observed_diagonal) == 0){
		  
#ifdef STATS
		  f->nb_chains_built++;
#endif
		  
		  /* create at tuple list element */
		  f->last_tl->next = CreateTupleList();
		  f->last_tl = f->last_tl->next;
		  
		  /* create a tuple [j] and update lists */
		  t  = CreateTuple(j,observed_diagonal,seed);
		  
		  
		  f->last_tl->first_tuple = t;
		  PUT_TAB(last_tuple_ptr_with_diag,observed_diagonal,t);
		}
		else
		  /* (2) we search for the last tuple [j] instanciation 
		   */
		  t = (tuple *)GET_TAB(last_tuple_ptr_with_diag,observed_diagonal) ;
		
		
		/*
		 * (3) iterate tuples and add this new one to the most adapted one
		 */
		
		do{
		  
		  if (  diagonal == t->diagonal  && f->i_current - t->occurrence <= gp_seeds_span[seed] + 1){
		    
		    /* (3a) possible fusion with this tuple (no instanciation)
		     *	
		     * [inter-tuple overlap]	
		     */      
		    
		    t->leftsize          +=  f->i_current -  t->occurrence;
		    t->lastSeed           = seed;
		    t->bitweight         += gp_automata[(~(t->overlappedhits<<(f->i_current -  t->occurrence)))&gp_seeds_bitweight[t->lastSeed]];
		    t->overlappedhits     = t->overlappedhits<<(f->i_current -  t->occurrence) | gp_seeds_bitweight[t->lastSeed];
		    t->overlappedhits    &= AUTOMATAMASK; 
		    t->occurrence         = f->i_current;
		    goto no_maj_last_tuple_ptr_with_diag;
		    
		  }
		  
		  /*	parcours de la liste */
		  if(!(t->next))
		    break;
		  t = t->next;
		  
		}while(1);
		
		/*
		 * (3d) any fusion has not been possible
		 *
		 * we instanciate the tuple and add this one at the end of the chain.
		 *
		 */
		
		t_ = CreateTuple(f->i_current, diagonal,seed);
		t->next = t_;
		PUT_TAB    (last_tuple_ptr_with_diag,diagonal,t_);
		
		/*
		 * (4) dont reset last_tuple_ptr_with_diag[diagonal] to NULL 
		 * because it has been modified in (3)
		 */
		goto no_maj_last_tuple_ptr_with_diag;
		
	      }
	      
	      /*---------------------------------------------------------------------*/
	      
	      
	      /* diagonal walk d,d+1,d-1,... */
	      delta_diagonal ++;
	      observed_diagonal = diagonal + gv_delta_shift[delta_diagonal];
	      observed_diagonal = MAX(MIN(observed_diagonal,keysize_query-1),1);
	      /* border effects avoided*/
	      
	    }/*  while : diagonal walk */
	  
	  
	  /* reset last tuple instancied to NULL (because not instance is created) 
	   */ 
	  DEL_TAB (last_tuple_ptr_with_diag,diagonal,NULL);	
	  
	no_maj_last_tuple_ptr_with_diag:
	  
	  /* 
	   * update last tuple position  with diagonal = 'diagonal'
	   */
#ifdef STATS
	  f->nb_seeds++;
#endif
	  PUT_TAB(last_tuple_pos_with_diag,diagonal,f->i_current);
#ifdef CACHE
	  last_tuple_pos_with_diag_hach[H_DIAG(diagonal)]=f->i_current;
#endif
	  /* iterate 'w' code list before f->i_current position */
#ifdef KEYLISTCOMPRESS		
	  index ++;
	  i_previous = keylist[seed][index]; 	
#else	
	  i_previous = index = keylist[seed][index];
#endif
	} /* while i_previous */
      }
    } /* for each seed */
    
	
	
	
    if (!(f->i_current % ALIGN_EVERY_NB_ITER)){
#ifdef TRACE
      Display_Progress((unsigned)f->i_current,(unsigned)(keysize_query - 1),f);
#endif
#ifdef STATS
      f->clock_chain += clock() - last_clock;
      last_clock=  clock();
#endif		
#ifdef DEBUG_ASSEMBLE
      DisplayListTupleList(f->first_tl);
#endif
      AlignAndFree(
		   datarev, datasize,
		   data,    datasize,
		   0,
		   f
		   );
	  
#ifdef STATS
      f->clock_align += clock() - last_clock;
      last_clock=  clock();
#endif
  
    }
  }/*end of [3] */
  
  
#ifdef TRACE
  Display_Progress((unsigned)f->i_current,(unsigned)(keysize_query - 1),f);
#endif  
  
#ifdef STATS
  f->clock_chain += clock() - last_clock;
  last_clock=  clock();
#endif	
  
  /* on aligne les tuples restants */
#ifdef DEBUG_ASSEMBLE
  DisplayListTupleList(f->first_tl);
#endif
  AlignAndFree(
	       datarev,    datasize,
	       data, datasize,
	       1,
	       f);
#ifdef STATS	
  f->clock_align += clock() - last_clock;
  last_clock =  clock();
#endif
  
#ifdef CACHE
 FREE(last_tuple_pos_with_diag_hach ,((MEMOPT_TABSIZE)/DIV_HACH+1) * sizeof(int));
#endif

  FREE_TAB(last_tuple_pos_with_diag);
  FREE_TAB(last_tuple_ptr_with_diag);
  
  for(seed=0;seed<gp_nb_seeds;seed++) {
    FREE(keylist[seed], keylist_size[seed]);
    FREE(first[seed],   first_size[seed]);
  }
  return  0;
}




int MultiAssemble_Double(char * data_query, int datasize_query,
			 char * data_text,  int datasize_text, 
			 int nbchunks_text, char ** chunkname_text,
			 int * chunksize_text,
			 int * chunkstrt_text,
			 Feature  *f ){
  /* used for chunk selection */
  int i_base                     = 0;
  int max_chunk_datasize_text    = 0;
  int keysize_text               = 0;
  
  int     chunk_keysize_text     = 0;
  char *  chunk_data_text        = data_text;
  
  /*  main variables 
   *  for the linked list computation
   */
  
  int * keylist_query[MAX_SEED];
  int   keylist_query_size[MAX_SEED];
  int * first_query  [MAX_SEED];
  int   first_query_size[MAX_SEED];
  
  int   keysize_query = datasize_query - gp_seeds_span_min + 1;

#ifdef MEMOPT
  int   memopt_modmask = 0;
#endif

  /* for the linking algorithm */
  int i_previous;
  int index,index_fin;
  int j;
  int diagonal,observed_diagonal,delta_diagonal;
  
  
  /* auxiliary variables */
  int      code;
#ifdef LOW_MEMORY
  /* use hash table*/
  Table_hash  last_tuple_pos_with_diag;
  Table_hash  last_tuple_ptr_with_diag;
#else
  int  *   last_tuple_pos_with_diag = NULL;
  tuple ** last_tuple_ptr_with_diag = NULL;
  /* >> */
  int  *   last_chunk_diag_used     = NULL;
  int      nb_last_chunk_diag_used  = 0;
  /* << */
#endif 
  
  /* sequence covering variables */ 
  tuple     * t, * t_;
  
  /* Multi seed */
  int seed;
  
  /* Hach table */
#ifdef CACHE
  int j_hack,bcode; 
  register int borne_min_hash=gp_rho_stat/2-1;
  register int borne_max_hash=DIV_HACH-gp_rho_stat/2+1;
  int *last_tuple_pos_with_diag_hach;
#endif
  
  /* to keep lists of tuple found */
  f->first_tl = CreateTupleList();
  f->last_tl  = f->first_tl;
  
  { register int i_chunk;
  for(i_chunk = 0 ; i_chunk < nbchunks_text; i_chunk++)
    max_chunk_datasize_text = MAX(max_chunk_datasize_text, chunksize_text[i_chunk]);
  }
  keysize_text = max_chunk_datasize_text - gp_seeds_span_min + 1; 
  
  
  /* [0] left correction factor set to keysize_query */
  f->left_correction =  keysize_query;
  MEMOPT_SETMOD;
  
  /* [1] key linked list creation */
#ifdef STATS
  last_clock = clock();
#endif
  for(seed=0;seed<gp_nb_seeds;seed++) {
    CreateKeyList(data_query,datasize_query,seed,&first_query[seed],&first_query_size[seed],&keylist_query[seed],&keylist_query_size[seed],f);
  }
  
  /* [2] allocations */
  INIT_TAB(last_tuple_pos_with_diag,int *,MEMOPT_TABSIZE * sizeof(int),0xff);
  INIT_TAB(last_tuple_ptr_with_diag,tuple **,MEMOPT_TABSIZE * sizeof(int),0x00);
  
#ifdef CACHE 
  last_tuple_pos_with_diag_hach = (int *) MALLOC((unsigned)(((MEMOPT_TABSIZE)/DIV_HACH+1) * sizeof(int)));
  ASSERT(last_tuple_pos_with_diag_hach,"Assemble");
  memset(last_tuple_pos_with_diag_hach,0xff,(unsigned)(((MEMOPT_TABSIZE)/DIV_HACH+1) * sizeof(int)));
#endif 
  
#ifdef STATS
  f->clock_pre += clock() - last_clock;
  last_clock = clock();
#endif
  

/* >> */
#ifndef LOW_MEMORY
  last_chunk_diag_used  = (int *) MALLOC((unsigned)LAST_CHUNK_DIAG_NB * sizeof(int));
  ASSERT(last_chunk_diag_used,"Assemble");
  memset(last_chunk_diag_used,0x00,((unsigned)LAST_CHUNK_DIAG_NB * sizeof(int)));
#endif 
/* << */


  /* for each chunk element */
  for( f->i_chunk = 0 ; f->i_chunk < nbchunks_text ; f->i_chunk++ ){
    
    /*  
     * [3] clean used tables  and
     *     and set parameters  
     */

    
    
    chunk_keysize_text  =  chunksize_text[f->i_chunk] - gp_seeds_span_min + 1;
    chunk_data_text    +=  ((f->i_chunk == 0) ? (0) : (chunksize_text[f->i_chunk-1]));
    i_base             +=  ((f->i_chunk == 0) ? (0) : (chunksize_text[f->i_chunk-1])); 
    
    /* >> */
#ifdef LOW_MEMORY	
    RESET_TAB(last_tuple_pos_with_diag,MEMOPT_TABSIZE*sizeof(int),0xff);
    RESET_TAB(last_tuple_ptr_with_diag,MEMOPT_TABSIZE*sizeof(int),0x00);
#else
    if (nb_last_chunk_diag_used < LAST_CHUNK_DIAG_NB) {
      while (nb_last_chunk_diag_used > 0) {
	nb_last_chunk_diag_used--;
	DEL_TAB(last_tuple_pos_with_diag,last_chunk_diag_used[nb_last_chunk_diag_used],-1);
	DEL_TAB(last_tuple_ptr_with_diag,last_chunk_diag_used[nb_last_chunk_diag_used],NULL);
      }
    } else {
      nb_last_chunk_diag_used = 0;
      RESET_TAB(last_tuple_pos_with_diag,MEMOPT_TABSIZE*sizeof(int),0xff);
      RESET_TAB(last_tuple_ptr_with_diag,MEMOPT_TABSIZE*sizeof(int),0x00);
    }
#endif
    /* << */


    /* [4] algorithme de creation des tuples chains */
#ifdef DEBUG_ASSEMBLE
    fprintf(stdout,"> f->i_chunk : %d, chunk_keysize_text = %d \n",f->i_chunk,chunk_keysize_text);fflush(NULL);
#endif
    /* for each pos  */
    for(f->i_current = 0 ; f->i_current < chunk_keysize_text ; f->i_current ++ ) {
      /* for each seed */
      for( seed = 0 ; seed < gp_nb_seeds ; seed++) {
	if (  f->i_current < chunksize_text[f->i_chunk] - gp_seeds_span[seed] + 1 ) {
	    
	  KEY(code,chunk_data_text,f->i_current,seed);    /* we take the code of the kmot 'w' located at f->i_current position on the "Text" */

#ifdef KEYLISTCOMPRESS	
	  index      = first_query[seed][code];  /* search the first occurrence of 'w' on the sequence */
	  index_fin  = first_query[seed][code+1];
	  i_previous = keylist_query[seed][index];
#else		
	  index      = first_query[seed][code];  /* search the first occurrence of 'w' on the sequence */
	  index_fin  = -1;
	  i_previous = index;
#endif

	  /* for each occurrence of 'w' on the "Query" sequence */
	  while(index != index_fin){ 
	    
	    diagonal =  f->i_current - i_previous + keysize_query ;
	    
	    /* consider all the interesting diagonals by adding indel bias  0 +1 -1 +2... until we reach a new tuple or the upper indel bound */
	    observed_diagonal = diagonal;
	    delta_diagonal = 0;


#ifdef CACHE 
	  /* bcode is a hash of */
	  j_hack = last_tuple_pos_with_diag_hach [H_DIAG(observed_diagonal )]; 
	  bcode  = ((j_hack < f->i_current - gp_rho_stat) || (j_hack > f->i_current));
	  if ( observed_diagonal % DIV_HACH > borne_max_hash ) {
	    j_hack = last_tuple_pos_with_diag_hach [H_DIAG(observed_diagonal)+1]; 
	    bcode  = bcode && ((j_hack< f->i_current - gp_rho_stat)||(j_hack > f->i_current)); 
	  }
	  if (observed_diagonal % DIV_HACH < borne_min_hash) {
	    j_hack = last_tuple_pos_with_diag_hach [H_DIAG(observed_diagonal)-1]; 
	    bcode = bcode && ((j_hack< f->i_current - gp_rho_stat)||(j_hack > f->i_current));
	  } 
	  if (bcode==0)
#endif
	    while( delta_diagonal < 2 * gp_delta_stat + 1 ){
		
		/* search for interesting tuples */
		j =  GET_TAB_MIN(last_tuple_pos_with_diag,observed_diagonal,f->i_current - gp_rho_stat);
#ifdef DEBUG_ASSEMBLE
		{ int i;
		  for(i = 0 ; i < keysize_query+keysize_text ; i++ ) {
		    printf("[%d]",last_tuple_pos_with_diag[i]);
		}
		printf("\n");
		}
		printf(" %d < observed_diagonal = %d <= %d\n",f->i_current , observed_diagonal , f->i_current+keysize_query);
		assert(f->i_current<observed_diagonal);
		assert(observed_diagonal<=f->i_current+keysize_query);
		printf("j = %x\n",j);
#endif
		/* if tuples found are respecting criteria (neighboors on diagonal and on sequences) */
		if (j > f->i_current -  gp_rho_stat                    && 
		    j <= f->i_current                                  &&
		    j - observed_diagonal <= f->i_current - diagonal   && 
		    j != -1 ) { 
#ifdef DEBUG_ASSEMBLE
		  printf("j ok, last_tuple_ptr_with_diag[%d] = %x\n",
			 observed_diagonal,
			 (int)last_tuple_ptr_with_diag[observed_diagonal]);
#endif

		  /*---------------------------------------------------------------------*/
		  
		  /* (1) possibly create one instance of tuple [j] 
		   * if this one was not in any list. 
		   */
		  if (EXIST_TAB(last_tuple_ptr_with_diag,observed_diagonal) == 0){
#ifdef STATS
		    f->nb_chains_built++;
#endif
		    /* create at tuple list element */
		    f->last_tl->next = CreateTupleList();
		    f->last_tl = f->last_tl->next;
		    
		    /* create a tuple [j] and update lists */
		    t = CreateTuple(j,observed_diagonal,seed);
		    f->last_tl->first_tuple = t;
		    PUT_TAB(last_tuple_ptr_with_diag,observed_diagonal,t); 
		  }
		  else
		    
		    
		    /* (2) we search for the last tuple [j] instanciation 
		     */
		    t = (tuple *)GET_TAB(last_tuple_ptr_with_diag,observed_diagonal) ;
		  
		  
		  /*
		   * (3) iterate tuples and add this new one to the most adapted one
		   */
		  
		  do{
		    
		    if (  diagonal == t->diagonal  && f->i_current - t->occurrence <= gp_seeds_span[seed] + 1){
		      
		      /* (3a) possible fusion with this tuple (no instanciation)
		       *	
		       * [inter-tuple overlap]	
		       */ 
		      
		      t->leftsize          +=  f->i_current -  t->occurrence;
		      t->lastSeed           = seed;
		      t->bitweight         += gp_automata[(~(t->overlappedhits<<(f->i_current -  t->occurrence)))&gp_seeds_bitweight[t->lastSeed]];
		      t->overlappedhits     = t->overlappedhits<<(f->i_current -  t->occurrence) | gp_seeds_bitweight[t->lastSeed];
		      t->overlappedhits    &= AUTOMATAMASK;
		      t->occurrence         = f->i_current;		
		      goto no_maj_last_tuple_ptr_with_diag;
		      
		    }
		    
		    
		    /*	parcours de la liste */
		    if(!(t->next))
		      break;
		    t = t->next;
		    
		  }while(1);
		  
		  /*
		   * (3d) any fusion has not been possible
		   *
		   * we instanciate the tuple and add this one at the end of the chain.
		   *
		   */
		  
		  t_ = CreateTuple(f->i_current, diagonal,seed);
		  t->next = t_;
		  PUT_TAB(last_tuple_ptr_with_diag,diagonal,t_);
		  
		  /*
		   * (4) dont reset last_tuple_ptr_with_diag[diagonal] to NULL
		   * because it has been modified in (3)
		   */
		  goto no_maj_last_tuple_ptr_with_diag;
		  
		}
		
		/*---------------------------------------------------------------------*/
		
		
		/* diagonal walk d,d+1,d-1,... */
		delta_diagonal ++;
		observed_diagonal = diagonal + gv_delta_shift[delta_diagonal];
		observed_diagonal = MAX(MIN(observed_diagonal,f->i_current+keysize_query),f->i_current+1);
		/* border effects avoided*/
		
	      }/*  while : diagonal walk */
	    
	    
	    /* reset last tuple instancied to NULL (because not instance is created)
	     */ 
	    
	    DEL_TAB(last_tuple_ptr_with_diag,diagonal,NULL);	
	    
	    
	  no_maj_last_tuple_ptr_with_diag:
	    
	    /* 
	     * update last tuple position  with diagonal = 'diagonal' 
	     */
#ifdef STATS
	    f->nb_seeds++;
#endif
	    PUT_TAB(last_tuple_pos_with_diag,diagonal,f->i_current);
 /* >> */
#ifndef LOW_MEMORY
	    if ( nb_last_chunk_diag_used < LAST_CHUNK_DIAG_NB)
	      last_chunk_diag_used[nb_last_chunk_diag_used++] = diagonal;
#endif
 /* << */   

#ifdef CACHE
	    last_tuple_pos_with_diag_hach[H_DIAG(diagonal)] = f->i_current;
#endif
	    /* iterate 'w' code list before f->i_current position */
#ifdef KEYLISTCOMPRESS		
	    index ++;
	    i_previous = keylist_query[seed][index]; 	
#else	
	    i_previous = index = keylist_query[seed][index];
#endif
	  }
	}
      }/* for each seed */
      
      
      
      if (!(f->i_current % ALIGN_EVERY_NB_ITER)){
#ifdef TRACE
	Display_Progress((unsigned)f->i_current+i_base,(unsigned)(datasize_text-1),f);    
#endif
#ifdef STATS
	f->clock_chain += clock() - last_clock;
	last_clock=  clock();
#endif
#ifdef DEBUG_ASSEMBLE
	DisplayListTupleList(f->first_tl);
#endif	
	AlignAndFree(
		     data_query,        datasize_query,
		     chunk_data_text,   chunksize_text[f->i_chunk], 
		     0,
		     f);
#ifdef STATS
	f->clock_align += clock() - last_clock;
	last_clock   = clock();
#endif
	
      }
    }/*end of  [3] (keysize_text loop)  */
#ifdef TRACE
    Display_Progress((unsigned)f->i_current+i_base,(unsigned)(datasize_text),f);   
#endif
#ifdef STATS
    f->clock_chain += clock() - last_clock;
    last_clock=  clock();
#endif	
#ifdef DEBUG_ASSEMBLE
    DisplayListTupleList(f->first_tl);
#endif	
    
    AlignAndFree(
		 data_query     ,  datasize_query,
		 chunk_data_text,  chunksize_text[f->i_chunk], 
		 1,
		 f);
    
#ifdef STATS		
    f->clock_align += clock() - last_clock;
    last_clock =  clock();
#endif   	
    
  }/*end of  [4] (chunk loop) */
  
/* >> */
#ifndef LOW_MEMORY
  FREE(last_chunk_diag_used,(unsigned)LAST_CHUNK_DIAG_NB * sizeof(int));
#endif
/* << */   


#ifdef CACHE
  FREE(last_tuple_pos_with_diag_hach ,((MEMOPT_TABSIZE)/DIV_HACH+1) * sizeof(int));
#endif

  FREE_TAB(last_tuple_pos_with_diag);
  FREE_TAB(last_tuple_ptr_with_diag);

  for(seed=0;seed<gp_nb_seeds;seed++) {
    FREE(keylist_query[seed], keylist_query_size[seed]);
    FREE(first_query[seed],   first_query_size[seed]);
  }   
  return 0;
}



